@@ -11,9 +11,7 @@ our @ISA = qw(Exporter);
 our @EXPORT_OK = qw($JCODE_COMPAT $VERSION);
 
 our @EXPORT = qw(mime_eco);
-our $VERSION = '0.34';
-
-our $JCODE_COMPAT = 0; # compatible with Jcode
+our $VERSION = '0.50';
 
 use constant HEAD   => '=?UTF-8?B?';
 use constant HEAD_J => '=?ISO-2022-JP?B?';
@@ -24,106 +22,225 @@ sub mime_eco {
     my $charset = shift || 'UTF-8';
     my $lf  = shift || "\n";
     my $bpl = shift || 76;
+    my $mode = shift;
+    $mode = 2 unless defined $mode;
+    my $csn;
     my $pos = 0;
     my $np;
     my $refsub;
 
-    my @words;
-    my $wpart = '';
-
+    my ($w1, $w1_len, $w2);
+    my ($sps, $sps_len);
+    my $sp0;
+    my $sp1 = '';
     my $result = '';
-    my $word_len;
-    my $word1 = 0;
+    my $ascii = 1;
     my $tmp;
+    my $count = 0;
 
     return '' unless defined $str;
-    return '' if $str eq '';
+    return $str if $str =~ /^\s*$/;
 
     if ($charset eq 'UTF-8') {
-	$refsub = \&add_enc_word_utf8;
+	$csn = 1;
     }
     elsif ($charset eq 'ISO-2022-JP') {
-	$refsub = \&add_enc_word_7bit_jis;
+	$csn = 0;
     }
     else {
 	return undef;
     }
     my ($trailing_crlf) = ($str =~ /(\n|\r|\x0d\x0a)$/o);
 
-    for my $word (split /\s+/, $str) {
-	if ($word =~ /[^\x21-\x7e]/) {
-	    if ($wpart eq '') {
-		$wpart = $word;
-	    }
-	    else {
-		$wpart .= " $word";
-	    }
-	}
-	else {
-	    if ($wpart ne '') {
-		push(@words, $wpart);
-		$wpart = '';
-	    }
-	    push(@words, $word);
-	}
-    }
-    push(@words, $wpart) if $wpart ne '';
+    $str =~ tr/\n\r//d;
 
-    for my $word (@words) {
-	if ($word eq '') {
-	    $word1 = 1;
-	    next;
+    $str =~ /(\s*)(\S+)/gc;
+    ($sps, $w2) = ($1, $2);
+    $ascii = 0 if $w2 =~ /[^\x21-\x7e]/;
+    $w1 = "$sps$w2";
+
+    if ($mode == 2) {
+	$mode = ($w1 =~ /^(?:Subject:|Comments:)$/i) ? 0 : 1;
+    }
+    if ($csn == 1) {
+	$refsub = ($mode == 1) ?
+	    \&add_enc_word_sth : \&add_enc_word_utf8;
+    }
+    else {
+	$refsub = ($mode == 1) ?
+            \&add_enc_word_sth : \&add_enc_word_7bit_jis;
+    }
+
+    while ($str =~ /(\s*)(\S+)/gc) {
+	($sps, $w2) = ($1, $2);
+
+	if ($w2 =~ /[^\x21-\x7e]/) {
+	    if ($ascii) { # "ASCII \s+ WCHAR"
+		$sp0 = chop($sps);
+		$w1 .= $sps;
+		$w1_len = length($w1);
+		if ($count == 0) {
+		    $result = $w1;
+		    $pos = $w1_len;
+		}
+		else {
+		    if (($count > 1) and ($pos + $w1_len + 1 > $bpl)) {
+			$result .= "$lf$sp1$w1";
+			$pos = $w1_len + 1;
+		    }
+		    else {
+			$result .= "$sp1$w1";
+			$pos += $w1_len + 1;
+		    }
+		}
+		$sp1 = $sp0;
+		$w1 = $w2;
+	    }
+	    else { # "WCHAR \s+ WCHAR"
+		$w1 .= "$sps$w2";
+	    }
+	    $ascii = 0;
 	}
-	if ($word =~ /[^\x21-\x7e]/) {
-	    if ($pos == 0 and $word1 == 0) {
-		$tmp = &$refsub($word, 0, $lf, $bpl, \$np);
+	else { # "ASCII \s+ ASCII" or "WCHAR \s+ ASCII"
+	    $w1_len = length($w1);
+	    if ($ascii) { # "ASCII \s+ ASCII"
+		if ($count == 0) {
+                    $result = $w1;
+                    $pos = $w1_len;
+                }
+		else {
+		    if (($count > 1) and ($pos + $w1_len + 1 > $bpl)) {
+                        $result .= "$lf$sp1$w1";
+                        $pos = $w1_len + 1;
+                    }
+                    else {
+                        $result .= "$sp1$w1";
+                        $pos += $w1_len + 1;
+                    }
+		}
+	    }
+	    else { # "WCHAR \s+ ASCII"
+		if ($count == 0) {
+                    $result = &$refsub($w1, 0, $lf, $bpl, \$np, 0, $csn);
+                    $pos = $np;
+                }
+		else {
+		    $tmp =
+			&$refsub($w1, 1 + $pos, $lf, $bpl, \$np, 0, $csn);
+
+		    $result .= ($tmp =~ s/^ /$sp1/) ? "$lf$tmp" : "$sp1$tmp";
+		    $pos = $np;
+		}
+	    }
+	    $sps_len = length($sps);
+
+	    if ($pos >= $bpl) {
+		$sp0 = substr($sps, 0, 1);
+		$w2 = substr($sps, 1) . $w2;
+	    }
+	    elsif ($pos + $sps_len - 1 > $bpl) {
+		$result .= substr($sps, 0, $bpl - $pos);
+		$sp0 = substr($sps, $bpl - $pos, 1);
+		$w2 = substr($sps, $bpl - $pos + 1) . $w2;
+		$pos = $bpl;
 	    }
 	    else {
-		$tmp =
-		    &$refsub($word, 1 +
-				    ($JCODE_COMPAT ? length($result) : $pos),
-				    $lf, $bpl, \$np);
-		$result .= ($tmp =~ /^ /) ? $lf : ' ';
+		$sp0 = chop($sps);
+		$result .= $sps;
+		$pos += $sps_len - 1;
 	    }
-	    $result .= $tmp;
-	    $pos = $np;
+	    $sp1 = $sp0;
+	    $w1 = $w2;
+	    $ascii = 1;
+	}
+	$count++ if $count <= 1;
+    }
+    ($sps) = ($str =~ /(.*)/g); # All space of the remainder
+
+    $w1 .= $sps;
+    if ($ascii) {
+	if ($count == 0) {
+	    $result = $w1;
 	}
 	else {
-	    $word_len = length($word);
-	    if ($word_len > $bpl) {
-		$result .= "$lf $word";
-		$pos = $word_len + 1;
-		next;
-	    }
-	    if ($pos == 0 and $word1 == 0) {
-		$result = $word;
-		$pos = $word_len;
-		next;
-	    }
-	    if ($pos + $word_len + 1 > $bpl) {
-		$result .= "$lf $word";
-		$pos = $word_len + 1;
+	    $w1_len = length($w1);
+	    if (($count > 1) and ($pos + $w1_len + 1 > $bpl)) {
+		$result .= "$lf$sp1$w1";
 	    }
 	    else {
-		$result .= " $word";
-		$pos += $word_len + 1;
+		$result .= "$sp1$w1";
 	    }
 	}
     }
+    else {
+	if ($count == 0) {
+	    $result = &$refsub($w1, 0, $lf, $bpl, \$np, 0, $csn);
+	}
+	else {
+	    $tmp = &$refsub($w1, 1 + $pos, $lf, $bpl, \$np, 0, $csn);
+	    $result .= ($tmp =~ s/^ /$sp1/) ? "$lf$tmp" : "$sp1$tmp";
+	}
+    }
+
     return $trailing_crlf ? $result . $trailing_crlf : $result;
 }
 
 
-# add encorded-word for 7bit-jis string
+# add encorded-word (for structured header)
 #   parameters:
 #     str : 7bit-jis string
 #     sp  : start position (indentation of the first line)
 #     lf  : line feed (default: "\n")
 #     bpl : bytes per line (default: 76)
 #     ep  : end position of last line (call by reference)
+#     rll : room of last line (default: 0)
+#     csn : charset number
+sub add_enc_word_sth {
+    my($str, $sp, $lf, $bpl, $ep, $rll, $csn) = @_;
+
+    my ($rb1, $rb2) = (0, 0); # '(' and ')' : round brackets
+    my $rb2s;
+    my $reg;
+    my $refsub;
+    my $tmp;
+
+    if ($csn == 1) {
+	$reg = qr/(\)\,?)$/;
+	$refsub = \&add_enc_word_utf8;
+    }
+    else {
+	$reg = qr/\e\(B[\x21-\x7e]*(\)\,?)$/;
+	$refsub = \&add_enc_word_7bit_jis;
+    }
+
+    if ($str =~ s/^\(//) {
+	$rb1 = 1;
+	$sp += 1;
+    }
+    if ($str =~ /$reg/) {
+	$rb2s = $1;
+	$rb2 = 1;
+	chop($str);
+	$rll = length($rb2s);
+	chop($str) if $rll > 1;
+    }
+    $tmp = &$refsub($str, $sp, $lf, $bpl, $ep, $rll, $csn);
+
+    if ($rb1) {
+	if ($tmp !~ s/^ / \(/) {
+	    $tmp = '(' . $tmp;
+	}
+    }
+    $tmp .= $rb2s if $rb2;
+
+    return $tmp;
+}
+
+
+# add encorded-word for 7bit-jis string
 sub add_enc_word_7bit_jis {
     require MIME::Base64;
-    my($str, $sp, $lf, $bpl, $ep) = @_;
+    my($str, $sp, $lf, $bpl, $ep, $rll, $csn) = @_;
 
     return '' if $str eq '';
 
@@ -138,10 +255,11 @@ sub add_enc_word_7bit_jis {
     my $result = '';
     my $str_pos;
     my $str_len = length($str);
+    my $ll_flag = 0;
 
     # encoded size + sp (18 is HEAD_J + TAIL)
     my $ep_tmp = int(($str_len + 2) / 3) * 4 + 18 + $sp;
-    if ($ep_tmp <= $bpl) {
+    if ($ep_tmp + $rll <= $bpl) {
 	$$ep = $ep_tmp;
 	return HEAD_J . MIME::Base64::encode_base64($str, '') . TAIL;
     }
@@ -204,15 +322,30 @@ sub add_enc_word_7bit_jis {
 
 	    # encoded size (19 is 18 + space)
 	    $ep_tmp = int(($str_len - $str_pos + $w_len + 2) / 3) * 4 + 19;
-	    if ($ep_tmp <= $bpl) {
+	    if ($ep_tmp + $rll <= $bpl) {
 		$chunk = $w . substr($str, $str_pos);
 		last;
 	    }
+	    elsif ($ep_tmp <= $bpl) {
+		$ll_flag = 1;
+	    }
             $chunk = $w;
             $chunk_len = $w_len;
             $sp = 1; # 1 is top space
         }
         else {
+	    if ($ll_flag and pos($str) == $str_len) {
+		if ($k_in_bak) {
+		    $chunk .= "\e\(B";
+		    $w = "\e$ec_bak" . $w;
+		    $w_len += 3;
+		}
+		$result .= HEAD_J .
+		    MIME::Base64::encode_base64($chunk, '') . TAIL . "$lf ";
+		$ep_tmp = int(($w_len + 2) / 3) * 4 + 19;
+		$chunk = $w;
+		last;
+	    }
             $chunk .= $w;
             $chunk_len += $w_len;
         }
@@ -228,7 +361,7 @@ sub add_enc_word_7bit_jis {
 # add encorded-word for utf8 string
 sub add_enc_word_utf8 {
     require MIME::Base64;
-    my($str, $sp, $lf, $bpl, $ep) = @_;
+    my($str, $sp, $lf, $bpl, $ep, $rll, $csn) = @_;
 
     return '' if $str eq '';
 
@@ -241,7 +374,7 @@ sub add_enc_word_utf8 {
 
     # encoded size + sp (12 is HEAD + TAIL)
     my $ep_tmp = int(($str_len + 2) / 3) * 4 + 12 + $sp;
-    if ($ep_tmp <= $bpl) {
+    if ($ep_tmp + $rll <= $bpl) {
 	$$ep = $ep_tmp;
 	return HEAD . MIME::Base64::encode_base64($str, '') . TAIL;
     }
@@ -267,11 +400,23 @@ sub add_enc_word_utf8 {
 
 	    # encoded size (13 is 12 + space)
             $ep_tmp = int(($str_len - $str_pos + 2) / 3) * 4 + 13;
-            if ($ep_tmp <= $bpl) {
+            if ($ep_tmp + $rll <= $bpl) {
 		utf8::encode($str); # UTF8 flag off
                 $chunk = substr($str, $str_pos);
                 last;
             }
+	    elsif ($ep_tmp <= $bpl) {
+		$w = chop($str);
+		utf8::encode($w); # UTF8 flag off
+		$w_len = length($w);
+		utf8::encode($str); # UTF8 flag off
+		$chunk = substr($str, $str_pos);
+		$result .= HEAD .
+		    MIME::Base64::encode_base64($chunk, '') . TAIL . "$lf ";
+		$ep_tmp = int(($w_len + 2) / 3) * 4 + 13;
+		$chunk = $w;
+		last;
+	    }
 	    $chunk = $w;
 	    $chunk_len = $w_len;
 	    $sp = 1; # 1 is top space
@@ -302,16 +447,16 @@ MIME::EcoEncode - MIME Encoding (Economical)
 
 This module implements RFC 2047 Mime Header Encoding.
 
-=head2 GLOBAL VARIABLES
-
-$MIME::EcoEncode::JCODE_COMPAT # compatible with Jcode (0 or 1)
-
 =head2 OPTIONS
 
-  $encoded = mime_eco($str, $charset, $lf, $bpl);
+  $encoded = mime_eco($str, $charset, $lf, $bpl, $mode);
                # $charset : 'UTF-8' or 'ISO-2022-JP'
                # $lf      : line feed (default: "\n")
                # $bpl     : bytes per line (default: 76)
+               # $mode    : 0 : unstructured header
+               #            1 : structured header
+               #            2 : auto (Subject or Comments ? 0 : 1)
+               #            (default: 2)
 
 =head1 SEE ALSO
 
